import geniz
from input import make_palindrome


@geniz.DebugAgent()
def test_make_palindrome():
    assert make_palindrome('') == ''
    assert make_palindrome('cat') == 'catac'
    assert make_palindrome('ai') == 'iai'
    assert make_palindrome('Hello') == 'HellolleH'
    assert make_palindrome('A!') == 'A!'
    assert make_palindrome('1#!') == '1#!1#'
    assert make_palindrome('hello world!') == '!ollehdlrowolleh'
    assert make_palindrome('a') == 'a'
    assert make_palindrome('aba') == 'ababa'
test_make_palindrome()


import geniz
from input import make_palindrome


@geniz.DebugAgent()
def test_make_palindrome():
    assert make_palindrome('') == ''
    print('Test with empty string passed.')
    assert make_palindrome('a') == 'aa'
    print('Test with single character string passed.')
    assert make_palindrome('racecar') == 'racecar'
    print('Test with a palindrome itself passed.')
    assert make_palindrome('ingoimm') == 'ginoming'
    print('Test with a non-palindrome string and its regular form passed.')
    assert make_palindrome('wel') == 'welwel'
    print('Test with simple concatenation case passed.')
    assert make_palindrome('he') == 'heh'
    print('Test with a palindrome of length 2 passed.')
    assert make_palindrome('d') == 'd'
    print('Test with single-character string resulting in palindrome upon append passed.')
    print('All tests passed.')
test_make_palindrome()


import ast
from typing import Optional

from pydantic import BaseModel


_CODE_AGENT_DECORATOR = ast.Call(
    func=ast.Attribute(
        value=ast.Name(id='sudocode', ctx=ast.Load()),
        attr='CodeAgent',
        ctx=ast.Load()),
    args=[],
    keywords=[])


_DEBUG_AGENT_DECORATOR = ast.Call(
    func=ast.Attribute(
        value=ast.Name(id='sudocode', ctx=ast.Load()),
        attr='DebugAgent',
        ctx=ast.Load()),
    args=[],
    keywords=[])


def maybe_add_decorators(tree: ast.Module, function_name: Optional[str] = None) -> bool:
    has_func_def = False
    tree.body.insert(0, ast.Import(names=[ast.alias(name='sudocode')]))
    for statement in reversed(tree.body):
        if isinstance(statement, ast.FunctionDef):
            if statement.name.startswith('test_'):
                statement.decorator_list = [_DEBUG_AGENT_DECORATOR]
            else:
                if function_name:
                    if statement.name == function_name:
                        has_func_def = True
                        statement.decorator_list = [_CODE_AGENT_DECORATOR]
    return has_func_def


def find_longest_code_block(text, function_name: Optional[str] = None, require_func_def = False):
    # May generate several code blocks, alwasy choose the longest.
    code_block_starters = text.split('```python')
    code_blocks = []
    for starter in code_block_starters:
        try:
            code_block = starter.split('```')[0]
            # Valid only if be able to parse it.
            # We assume only to generate function for now.
            tree = ast.parse(code_block)
            has_func_def = maybe_add_decorators(tree, function_name)
            if require_func_def:
                if not has_func_def:
                    continue
            code_block = ast.unparse(tree)
        except:
            continue
        code_blocks.append(code_block)
    return max(code_blocks, key=lambda x: len(x))


class PythonCode(BaseModel):
    # Code snippet.
    code_text: Optional[str] = None
    # Exception string if error from parsing
    parsing_error: Optional[str] = None

    def valid(self) -> bool:
        return self.code_text is not None

    @classmethod
    def extract_main_program_code_block_from_full_reply(
        cls,
        full_reply: str,
        function_name: Optional[str] = None
    ) -> "PythonCode":
        """Extract and validate the code block from replied texts."""
        try:
            code_gen_section = full_reply
            # May generate several code blocks, alwasy choose the longest.
            code_block_text = find_longest_code_block(
                code_gen_section, function_name=function_name, require_func_def=True)
            return cls(code_text=code_block_text)
        except Exception as e:
            return cls(code_text=None, parsing_error=str(e))

    @classmethod
    def extract_test_case_code_block_from_full_reply(
        cls,
        full_reply: str,
        function_name: Optional[str] = None
    ) -> "PythonCode":
        """Extract and validate test case code block from replied texts."""
        try:
            # May generate several code blocks, alwasy choose the longest.
            code_block_text = find_longest_code_block(
                full_reply, function_name=function_name)
            return cls(code_text=code_block_text)
        except Exception as e:
            return cls(code_text=None, parsing_error=str(e))


_TEST_CODE = '''
def make_palindrome(string: str) -> str:
    """ Find the shortest palindrome that begins with a supplied string.
    Algorithm idea is simple:
    - Find the longest postfix of supplied string that is a palindrome.
    - Append to the end of the string reverse of a string prefix that comes before the palindromic suffix.
    >>> make_palindrome('')
    ''
    >>> make_palindrome('cat')
    'catac'
    >>> make_palindrome('cata')
    'catac'
    """
    postfix = string[::-1]  # Reverse the string
    for i in range(len(string)):
        if string[:len(string)-i] == postfix[i:]:
            return string + string[:i][::-1]
    return ''

print(make_palindrome(""))
print(make_palindrome("cat"))
print(make_palindrome("cata"))
'''


def test_parse_and_add_decorator():
    tree = ast.parse(_TEST_CODE)
    maybe_add_decorators(tree)
    print(ast.unparse(tree))


if __name__ == '__main__':
    test_parse_and_add_decorator()



import traceback
from collections import defaultdict
from contextlib import contextmanager
from typing import Any, Dict, List, Tuple

from pydantic import BaseModel

from .util import shorten_answer

DATA_DIST = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

COLLECTION_MODE = False
CURRENT_COLLECTOR = None


class DataPoint(BaseModel):
    input: Tuple
    output: Any
    candidate: str = ''
    test_case: str = ''

    def debug_str(self):
        msg = f'DataPoint\n  input: {str(self.input)[:50]}\n  output: {str(self.output)[:50]}\n  candidate: {self.candidate}\n  test_case: {self.test_case}'
        return msg


def datapoint_to_input_output_str(datapoint):
    input_str = str(datapoint.input)
    if isinstance(datapoint.output, Exception):
        e = datapoint.output
        output_str = f'Exception({type(e).__name__}: {e})'
    else:
        output_str = str(datapoint.output)
    return input_str, output_str

class DataCollector(BaseModel):
    test_case: str
    data_points: List[DataPoint] = []

    def record(self, datapoint: DataPoint):
        # Skip invalid executions.
        # if datapoint.output is None or isinstance(datapoint.output, Exception):
        #     return
        datapoint.test_case = self.test_case
        self.data_points.append(datapoint)
        # print(f'Record {datapoint.debug_str()}')

    def merge(self):
        for datapoint in self.data_points:
            try:
                input_str, output_str = datapoint_to_input_output_str(datapoint)
                DATA_DIST[input_str][output_str][datapoint.candidate].append(
                    datapoint)
            except:
                pass


class ReenterException(Exception):
    pass


def _enter_collection_mode(collector: DataCollector):
    global COLLECTION_MODE, CURRENT_COLLECTOR
    if COLLECTION_MODE is True:
        raise ReenterException('re-enter collection mode')
    COLLECTION_MODE = True
    CURRENT_COLLECTOR = collector


def _leave_collection_mode():
    global COLLECTION_MODE, CURRENT_COLLECTOR
    if COLLECTION_MODE is False:
        print('Warning leave collection mode without entering')
    COLLECTION_MODE = False
    CURRENT_COLLECTOR = None


def is_data_collection_mode() -> bool:
    return COLLECTION_MODE


def get_data_collector() -> DataCollector:
    return CURRENT_COLLECTOR  # type: ignore


def reduce_to_most_frequent_answer() -> Dict[Tuple, str]:
    most_frequent = dict()
    for input, output_dist in DATA_DIST.items():
        if not output_dist:
            continue
        answer = max(output_dist.items(), key=lambda x: len(x[1]))
        most_frequent[input] = shorten_answer(answer[0])
    return most_frequent


def calculate_candidates_stats_scores(all_candidates: List[str] = []) -> Dict[str, int]:
    candidate_to_score = defaultdict(int)
    for c in all_candidates:
        candidate_to_score[c] = 0
    for _, output_dist in DATA_DIST.items():
        if not output_dist:
            continue
        sort_by_frequency = sorted(
            output_dist.items(), reverse=True, key=lambda x: len(x[1]))
        # Find the first valid output to credit candidates score.
        bounty = 3
        for _, candidate_dist in sort_by_frequency:
            for candidate in candidate_dist:
                candidate_to_score[candidate] += bounty
            bounty -= 1
            if bounty == 0:
                break
    return candidate_to_score


def get_test_dist() -> Dict[Tuple, Any]:
    test_dist = dict()
    for input, output_dist in DATA_DIST.items():
        dist = sorted([(len(points), output)
                      for output, points in output_dist.items()], reverse=True, key=lambda x: x[0])
        test_dist[input] = dist
    return test_dist


def get_candidate_input_output() -> Dict[str, Dict[str, DataPoint]]:
    candidate_input_output = defaultdict(dict)
    for input, output_dist in DATA_DIST.items():
        for _, candidate_dist in output_dist.items():
            for candidate_id, datapoints in candidate_dist.items():
                candidate_input_output[candidate_id][input] = datapoints[0]
    return candidate_input_output


def get_data_dist():
    return DATA_DIST


@contextmanager
def data_collection_mode(test_case):
    collector = DataCollector(test_case=test_case)
    try:
        _enter_collection_mode(collector)
        yield collector
    except Exception as e:
        print(f'  suppress exception: {e}')
        traceback.print_exc()
    finally:
        collector.merge()
        _leave_collection_mode()



import logging
import os
import pickle

import isort

from .llm import query_llm
from .python_code import PythonCode
from .util import alphanumeric_uuid, load_prompt

TEST_DESIGNER_PROMPT = load_prompt(
    os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "prompts/test_designer_prompt_simple.yaml",
    )
)


def create_test_file(source_filename, function_name, source_code) -> bool:
    source_error_filename = f'{source_filename}.error'
    module_name = source_filename.removesuffix('.py')
    new_id = str(alphanumeric_uuid()[-5:])
    output_filename = f'{module_name}_{new_id}_test.py'

    # Skip the submitted result.
    if source_filename == 'output.py':
        return False
    # Skip if original program already contains error
    # Maybe add another fixer here.
    if os.path.exists(source_error_filename):
        os.system(f'rm {source_error_filename}*')
        return False
    if os.path.exists(output_filename):
        return False

    logging.info(
        f'TestDesigner: generating {output_filename} for {function_name} in {source_filename}...')

    system_message = TEST_DESIGNER_PROMPT.format(
        function_name=function_name, input_code_snippet=source_code)
    message = f'''Please generate test cases for following code, and add "from {module_name} import {function_name}" in top and assume {function_name} is already defined in {module_name}.
```python
{source_code}
```

Test case is like:
```python
def test_{function_name}():
  assert xxxx
  assert xxxx
  ...
```
'''
    replys, histories = query_llm(message, system_message=system_message, filename=output_filename)
    for i, reply in enumerate(replys):
        test_code = PythonCode.extract_test_case_code_block_from_full_reply(
            reply, function_name=function_name)
        if not test_code.valid():
            logging.info(f'Test case is invalid:\n\n{test_code.code_text}')
            return False
        test_case_filename = f'{module_name}_{new_id}_{i}_test.py'
        with open(test_case_filename, 'w') as f:
            code_text = f'from {module_name} import {function_name}\n{test_code.code_text}'
            f.write(code_text)
            logging.info(f'Wrote to {test_case_filename}')
        isort.file(test_case_filename)
        new_history_filename = f'{test_case_filename}.history.pickle'
        with open(new_history_filename, 'wb') as f:
            pickle.dump(histories[i], f)
            logging.info(f'Wrote to {new_history_filename}')
    return True



# Mutation from 'input.py'
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np

import geniz


def expand_around_center(s: str, left: int, right: int) -> str:
    """Helper function to expand around a center and return the palindrome."""
    while left >= 0 and right < len(s) and (s[left] == s[right]):
        left -= 1
        right += 1
    return s[left + 1:right]

@geniz.CodeAgent()
def make_palindrome(s: str) -> str:
    """Find and return the shortest palindrome beginning with the given string."""
    if s == s[::-1]:
        return s + s
    for i in range(len(s)):
        palindrome = expand_around_center(s, i, i)
        palindrome2 = expand_around_center(s, i, i + 1)
        if len(palindrome) < len(palindrome2):
            result = s + palindrome[::-1]
        else:
            result = s + palindrome + palindrome[::-1]
        if is_valid_palindrome(result):
            return result
    raise ValueError('No palindrome could be adequately formed.')

def is_valid_palindrome(s: str) -> bool:
    """Helper to check if the constructed string s is a palindrome"""
    return s == s[::-1]
if __name__ == '__main__':
    print(make_palindrome(''))
    print(make_palindrome('cat'))
    print(make_palindrome('cata'))
    assert expand_around_center('a', 0, 1) == ''
    assert expand_around_center('a', 0, 0) == 'a'
    assert expand_around_center('ab', 0, 1) == 'b'
    assert expand_around_center('racecar', 0, 7) == 'racecar'
    assert expand_around_center('hello', 0, 4) == 'olleh'
    assert not expand_around_center('apple', 0, 2)
    assert is_valid_palindrome('aba') and (not is_valid_palindrome('abac'))
    assert is_valid_palindrome(make_palindrome('ab'))




import os

MODEL = os.getenv('AUTOGENESIS_MODEL', 'Phi-3-mini-128k-instruct-a100')
API_BASE = os.getenv('AUTOGENESIS_API_BASE', 'https://geniz.ai/v1')
API_KEY = os.getenv('AUTOGENESIS_API_KEY', 'sk-sudocode-ai')


import uuid
from datetime import datetime, timedelta

from pydantic import BaseModel

from .util import PersistStateToFile


class RoundInfo(BaseModel, PersistStateToFile):
    created: datetime = datetime.now()
    round: int = 1
    session_id: str = str(uuid.uuid4())

    def dead(self):
        if datetime.now() - self.created > timedelta(minutes=25):
            return True
        return False


def get_round_info():
    round_info = RoundInfo().load()
    return round_info

def get_round_filename_prefix():
    round_info = RoundInfo().load()
    return f'round_{round_info.round}'


import copy
import logging
import os
from functools import partial

import gradio as gr
import ray
from geniz.coder import (generate_code, generate_test,
                            get_test_and_candidate_info, load_locked_tests,
                            save_locked_tests)

logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S')

os.environ['RAY_IGNORE_UNHANDLED_ERRORS'] = '1'


if gr.NO_RELOAD:
    ray.init(include_dashboard=False, ignore_reinit_error=True)


def get_original_problem_prompt():
    with open('input.py', 'r') as f:
        code_file = f.read()
    return code_file


test_info, candidate_info, locked_tests = get_test_and_candidate_info()


_CSS = '''
.test_case_container {
    border-width: medium;
}

.test_case_locked {
    border-color: green;
}
'''

with gr.Blocks(css=_CSS) as demo:
    candidate_info_state = gr.State(candidate_info)
    test_info_state = gr.State(test_info)
    locked_tests_state = gr.State(locked_tests)

    def click_run_all_tests():
        test_info, candidate_info, locked_tests = get_test_and_candidate_info()
        return {
            candidate_info_state: candidate_info,
            test_info_state: test_info,
            locked_tests_state: locked_tests,
        }

    def click_gen_code():
        generate_code()
        test_info, candidate_info, locked_tests = get_test_and_candidate_info()
        return {
            candidate_info_state: candidate_info,
            test_info_state: test_info,
            locked_tests_state: locked_tests,
        }

    def click_gen_test():
        generate_test()
        test_info, candidate_info, locked_tests = get_test_and_candidate_info()
        return {
            candidate_info_state: candidate_info,
            test_info_state: test_info,
            locked_tests_state: locked_tests,
        }

    with gr.Row():
        with gr.Accordion(label='Problem Description', open=False):
            prompt_editor = gr.Code(
                value=get_original_problem_prompt,
                language='python',
                show_label=False,
                interactive=True,
            )
    gr.Markdown("---")
    with gr.Row():
        gen_code_button = gr.Button("Generate Code")
        gen_test_button = gr.Button("Generate Test")
        run_all_tests_button = gr.Button("Run All Tests")
    with gr.Row(equal_height=True):
        with gr.Column():
            @gr.render(inputs=[candidate_info_state])
            def render_candidate_data(input_0):
                for i, candidate_info in enumerate(input_0):
                    candidate_id = candidate_info['candidate_id']
                    candidate = candidate_info['candidate']
                    tests_score = candidate_info['tests_score']
                    stars = '⭐' * tests_score
                    with gr.Accordion(label=f'{candidate_id} {stars}',
                                      open=False,
                                      elem_id=candidate_id):
                        code_editor = gr.Code(
                            value=candidate.clean_source_code,
                            language='python',
                            interactive=True,
                            show_label=False)
                        with gr.Row():
                            delete_button = gr.Button('Delete', scale=0)
                            def click_delete_button(this_candidate_info, candidate_info):
                                this_candidate_id = this_candidate_info['candidate_id']
                                this_candidate = this_candidate_info['candidate']
                                this_candidate.delete()
                                return [c for c in candidate_info if c['candidate_id'] != this_candidate_id]
                            delete_button.click(partial(click_delete_button, copy.copy(candidate_info)),
                                                inputs=[candidate_info_state], outputs=[candidate_info_state])
        with gr.Column():
            @gr.render(inputs=[test_info_state])
            def render_test_data(input_0):
                if input_0 is None:
                    print('input_0 is None')
                    return
                for info in input_0:
                    default_output_str = info['default_output_str']
                    default_call_str = info['default_call_str']
                    locked = info['locked']
                    elem_id = info['id']
                    elem_classes = ['test_case_container']
                    if locked:
                        elem_classes.append('test_case_locked')
                    with gr.Group(elem_id=elem_id, elem_classes=elem_classes):
                        with gr.Row():
                            test_box = gr.Textbox(
                                default_call_str, show_label=False, interactive=True)
                        with gr.Row():
                            output_options = info['outputs']
                            output_radio_group = gr.Radio(
                                choices=output_options,
                                value=default_output_str,
                                container=False,
                                interactive=True,
                                label='Output Values')
                        lock_checkbox = gr.Checkbox(label='Lock', value=locked)

                        def output_radio_group_trigger(this_info, selected_output, locked_tests):
                            # TODO: color change for candidate boxes
                            output_info = this_info['outputs_info'].get(
                                selected_output, None)
                            if output_info is None or len(output_info) == 0:
                                return this_info['default_call_str'], locked_tests
                            if this_info['input'] in locked_tests:
                                if selected_output != locked_tests[this_info['input']]:
                                    locked_tests[this_info['input']] = selected_output
                                    save_locked_tests(locked_tests)
                            return output_info[0]['call_str'], locked_tests

                        output_radio_group.change(
                            partial(output_radio_group_trigger, copy.copy(info)),
                            inputs=[output_radio_group, locked_tests_state],
                            outputs=[test_box, locked_tests_state])

                        def lock_checkbox_trigger(this_info, true_or_false, selected_output, candidate_info, locked_tests):
                            if true_or_false is True:
                                locked_tests[this_info['input']] = selected_output                            
                                save_locked_tests(locked_tests)
                                new_pass_candidates = [x['candidate_id'] for x in this_info['outputs_info'][selected_output]]
                                for info in candidate_info:
                                    if info['candidate_id'] in new_pass_candidates:
                                        info['passed_tests'].add(this_info['id'])
                                        info['tests_score'] = len(info['passed_tests'])
                            else:
                                locked_tests.pop(this_info['input'], None)                            
                                save_locked_tests(locked_tests)
                                pass_candidates_to_remove = [x['candidate_id'] for x in this_info['outputs_info'][selected_output]]
                                for info in candidate_info:
                                    if info['candidate_id'] in pass_candidates_to_remove:
                                        info['passed_tests'].remove(this_info['id'])
                                        info['tests_score'] = len(info['passed_tests'])    
                            return candidate_info, locked_tests

                        lock_checkbox.change(
                            partial(lock_checkbox_trigger, copy.copy(info)),
                            inputs=[lock_checkbox, output_radio_group, candidate_info_state, locked_tests_state],
                            outputs=[candidate_info_state, locked_tests_state],
                            js='''(x, y, z, p) => {
    var element = document.getElementById("''' + str(elem_id) + '''");
    if (x) {
        element.classList.add("test_case_locked");
    } else {
        element.classList.remove("test_case_locked");
    }
    return [x, y, z, p];
}
''')

    gen_code_button.click(click_gen_code, inputs=None, outputs=[
                          candidate_info_state, test_info_state, locked_tests_state])
    gen_test_button.click(click_gen_test, inputs=None, outputs=[
                          candidate_info_state, test_info_state, locked_tests_state])
    run_all_tests_button.click(
        click_run_all_tests, inputs=None, outputs=[candidate_info_state, test_info_state, locked_tests_state])
    # code_editor.change(code_editor_change, code_editor, None)
    # gr.on(triggers=None, fn=click_run_all_tests, inputs=[], every=2)
    # dep = demo.load(click_run_all_tests, inputs=[], outputs=[test_info_state], every=2)

if __name__ == "__main__":
    demo.launch(debug=True)



# Mutation from 'input.py'
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np

import geniz


@geniz.CodeAgent()
def make_palindrome(s: str) -> str:
    """
    Finds the shortest palindrome that includes the input string. If the string is already a palindrome,
    a new string made with a copy of the original string suffices.
    The solution iteratively builds up the palindrome, attempting the smallest part of the string first,
    followed by appending the reverse of its suffix if needed.

    :param s: The input string.
    :return: The shortest palindrome constructed from the input string.
    """
    if s == s[::-1]:
        return s
    for i in range(len(s)):
        if s[:len(s) - i] == s[:len(s) - i][::-1]:
            if not s[len(s) - i:] or s[:len(s) - i] == s[len(s) - i:][::-1]:
                return s + s[:len(s) - i][::-1]
            else:
                continue
print(make_palindrome(''))
print(make_palindrome('cat'))
print(make_palindrome('cata'))



import geniz

def is_palindrome(string: str) -> bool:
    """ Test if given string is a palindrome """
    return string == string[::-1]


@geniz.CodeAgent()
def make_palindrome(string: str) -> str:
    """ Find the shortest palindrome that begins with a supplied string.
    Algorithm idea is simple:
    - Find the longest postfix of supplied string that is a palindrome.
    - Append to the end of the string reverse of a string prefix that comes before the palindromic suffix.
    >>> make_palindrome('')
    ''
    >>> make_palindrome('cat')
    'catac'
    >>> make_palindrome('cata')
    'catac'
    """



# Mutation from 'input.py'
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np

import geniz


@geniz.CodeAgent()
def make_palindrome(s):
    """
    Generates the shortest palindrome by appending the reverse of the prefix to the given string with its longest palindromic suffix.
    
    Parameters:
    s (str): The input string.
    
    Returns:
    str: The resulting shortest palindrome.
    
    Example:
    >>> make_palindrome('')
    ''
    >>> make_palindrome('cat')
    'catac'
    """
    rev_tail = s[::-1]
    for i in range(len(s), 0, -1):
        if rev_tail[:i] == rev_tail[i - 1::-1]:
            palindromic_suffix = rev_tail[:i]
            remaining_prefix = s[len(palindromic_suffix):]
            return palindromic_suffix + remaining_prefix[::-1]
if __name__ == '__main__':
    print(make_palindrome(''))
    print(make_palindrome('cat'))
    print(make_palindrome('cata'))



# Mutation from 'input.py'
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np

import geniz


@geniz.CodeAgent()
def make_palindrome(s: str) -> str:
    """
    Transforms a given string into the shortest palindrome beginning with it.
    
    Algorithm Explanation:
    - Finds the longest palindromic suffix within the given string.
    - Attaches the reverse of the prefix preceding the palindromic suffix at the end, effectively 
      creating a palindrome that starts with the original string.
    
    The function efficiently works in O(n^2) time complexity due to the process to find the longest 
    palindromic suffix within the string. However, it optimally satisfies the problem statement.
    
    Parameters:
    s (str): The input string we transform into the specified palindrome format.
    
    Returns:
    str: The transformed string that is a palindrome resembling the original string with its mirrored portion at the end.
    
    Examples:
    >>> make_palindrome('')
    ''
    >>> make_palindrome('cat')
    'catac'
    >>> make_palindrome('blackcarrot')
    'blackcarrototcarbalc'
    """
    if not s:
        return ''
    is_palindrome = lambda x: x == x[::-1]
    for i in range(len(s) - 1, -1, -1):
        if is_palindrome(s[i:]):
            return s + s[:i][::-1]
if __name__ == '__main__':
    assert make_palindrome('abcd') == 'abcdcba'
    print('All test cases pass')



"""Code-agent test-bench

Pre-requisite:
> pip install -e .
"""

import argparse
import asyncio
import glob
import os
from collections import defaultdict
from pathlib import Path
import shutil
from subprocess import check_output
from typing import Optional

import ray
from datasets import load_dataset
from tqdm import tqdm
from utils.logging import PrintColor, format_color


os.environ['RAY_DEDUP_LOGS'] = '0'


def calculate_score(output: str):
    total = defaultdict(int)
    positive = defaultdict(int)

    overral_score = 0
    for output_file_path in glob.glob(f'{output}/**/output.pass', recursive=True):
        output_file = Path(output_file_path)
        if not output_file.exists():
            continue
        passed = int(output_file.read_text())
        world_folder = os.path.dirname(output_file)
        task_id_file = Path(os.path.join(world_folder, 'task_id.txt'))
        task_id = task_id_file.read_text()

        total[task_id] += 1
        positive[task_id] += passed

    for task_id, sum in sorted(total.items(), key=lambda x: int(x[0].split('/')[1])):
        pos = positive[task_id]
        if pos > 0:
            color = PrintColor.GREEN
            overral_score += 1
        else:
            color = PrintColor.RED
        print(format_color(
            f'{task_id}: {pos}/{sum} ({100*pos/sum:.1f}%)', color))
    print()
    if len(total) > 0:
        print(
            f'Overall score: {overral_score}/{len(total)} ({100*overral_score/len(total):.1f}%)')


def init_benchmark_scaffolding(world_folder: Path, entry, agent):
    task_id = entry['task_id']
    assert task_id.startswith('HumanEval/')
    prompt = entry['prompt']
    test = entry['test']
    entry_point = entry['entry_point']

    if agent in ['ningcoder', 'opencoder', 'baseline']:
        with open(os.path.join(world_folder, 'input.py'), 'w') as f:
            source_lines = prompt.split('\n')
            for i, line in enumerate(source_lines):
                if line.startswith(f'def {entry_point}'):
                    source_lines.insert(i, '@sudocode.CodeAgent()')
                    break
            final_prompt = '\n'.join(source_lines)
            final_prompt = 'import sudocode\n' + final_prompt
            f.write(final_prompt)
    else:
        with open(os.path.join(world_folder, 'input.txt'), 'w') as f:
            polished_prompt = f'{prompt}\n# The main function (entry point) is {entry_point}'
            f.write(polished_prompt)
    with open(os.path.join(world_folder, 'task_id.txt'), 'w') as f:
        f.write(task_id)

    eval_code = f'''
{prompt}

{test}

if __name__ == '__main__':
    try:
        import output
        check(output.final_answer)
        with open('output.pass', 'w') as f:
            f.write('1')
    except:
        import traceback
        with open('output.error', 'w') as f:
            f.write(traceback.format_exc())
        with open('output.pass', 'w') as f:
            f.write('0')
'''
    with open(os.path.join(world_folder, 'eval_output.py'), 'w') as f:
        f.write(eval_code)


@ray.remote
def run_eval_case(trial: str, entry, output: str, rerun_failures: bool = False, image: str = 'autogenesis', agent: str = 'ningcoder', model: str = 'Phi-3-mini-128k-instruct'):
    world_path = Path(os.path.join(output, trial))
    pass_file = Path(os.path.join(world_path, 'output.pass'))
    if pass_file.exists():
        passed = pass_file.read_text()
        if rerun_failures and passed != '1':
            shutil.rmtree(world_path, ignore_errors=True)
        else:
            return world_path, passed
    else:
        shutil.rmtree(world_path, ignore_errors=True)

    init_cmd = ['python', 'run.py',
                f'--world={world_path}', f'--agent={agent}', '--init_only']
    print(' '.join(init_cmd))
    check_output(init_cmd)

    init_benchmark_scaffolding(world_path, entry, agent)

    if agent in ['ningcoder']:
        max_rounds = 11
    if agent in ['opencoder']:
        max_rounds = 30
    else:
        max_rounds = 23

    run_cmd = ['python', 'run.py', f'--world={world_path}', f'--agent={agent}',
               f'--rounds={max_rounds}', '--save_log_to_file', f'--image={image}',
               f'--model={model}']
    print(' '.join(run_cmd))
    retry = 0
    while retry < 2:
        retry += 1
        try:
            check_output(run_cmd)
            if pass_file.exists():
                return world_path, pass_file.read_text()
        except KeyboardInterrupt:
            return world_path, '0'
        except:
            continue
    return world_path, '0'


def humaneval_bench(output: str,
                    max_rows: int = 200,
                    rerun_failures: bool = False,
                    hard: bool = False,
                    parallel: int = 1,
                    against_one_case: Optional[int] = None,
                    trials: int = 1,
                    image: str = 'autogenesis',
                    agent: str = 'ningcoder',
                    model: str = 'Phi-3-mini-128k-instruct'):
    calculate_score(output)

    dataset = load_dataset("openai_humaneval")
    dataset_len = len(dataset["test"])
    if against_one_case is not None:
        eval_cases = [against_one_case]
    else:
        if hard:
            # Average passing rate:
            #  10: 30%
            #  33: 30%
            #  64: 80%
            #  65:  0%
            #  83:  0%
            # 115: 80%
            # 130: 10%
            # 132:  0%
            # 145:  0%
            # 163:  0%
            eval_cases = [10, 33, 64, 65, 83, 115, 130, 132, 145, 163]
        else:
            eval_cases = list(range(dataset_len))
        eval_cases = eval_cases[:max_rows]

    eval_tasks = [(f'{case_id}_{trial}', dataset['test'][case_id], output, rerun_failures, image, agent, model)
                  for case_id in eval_cases for trial in range(trials)]

    running_refs = []
    for task in tqdm(eval_tasks, desc='Benchmarking...'):
        if len(running_refs) >= parallel:
            # update result_refs to only track the remaining tasks.
            ready_refs, running_refs = ray.wait(running_refs, num_returns=1)
            results = ray.get(ready_refs)
            if len(results) > 0:
                world_folder, passed = results[0]
                if passed == '1':
                    color = PrintColor.GREEN
                else:
                    color = PrintColor.RED
                print(format_color(
                    f'Finished: {passed} - {world_folder}', color))

        running_refs.append(run_eval_case.remote(*task))
    ray.get(running_refs)

    calculate_score(output)


async def mbpp_bench(output: str, max_rows: int = 10, rerun_failures: bool = False, hard: bool = False):
    raise RuntimeError('TODO: migrate to call Autogenesis framework')


def main():
    ray.init()

    parser = argparse.ArgumentParser(description="Benchmark code-gen agent")
    parser.add_argument("--dataset", type=str, default="humaneval",
                        help="The dataset to benchmark.")
    parser.add_argument("--max_rows", type=int, default=200,
                        help="Max rows to test")
    parser.add_argument("--output", type=str,
                        default="./output", help="Output folder")
    parser.add_argument("--rerun_failures", action='store_true',
                        help="Only re-run the failed cases")
    parser.add_argument("--hard", action='store_true',
                        help="Run a small subset of tests (if it exists) that tend to fail often. If we get these csases right, it's a great chance to beat SOTA.")
    parser.add_argument("--parallel", type=int, default=1,
                        help="Max concurrent running eval cases")
    parser.add_argument("--against_one_case", type=int, default=None,
                        help="case number. Test only one.")
    parser.add_argument("--trials", type=int, default=1,
                        help="How many independent trails for one case. (Test flakiness)")
    parser.add_argument("--image", type=str,
                        default="autogenesis", help="docker image")
    parser.add_argument("--agent", type=str, default="ningcoder",
                        help="The agent to use.")
    parser.add_argument("--model", type=str, default="Phi-3-mini-128k-instruct",
                        help="The LLM model.")
    args = parser.parse_args()

    os.makedirs(args.output, exist_ok=True)

    if args.dataset == 'humaneval':
        humaneval_bench(args.output, args.max_rows, args.rerun_failures,
                        args.hard, args.parallel, args.against_one_case, args.trials, args.image, args.agent,
                        args.model)
    if args.dataset == 'mbpp':
        asyncio.run(mbpp_bench(args.output + "_mbpp",
                    args.max_rows, args.rerun_failures, args.hard))


if __name__ == '__main__':
    main()



import glob
import importlib
from os.path import basename, dirname, isfile, join

def auto_import():
    modules = glob.glob(join(dirname(__file__), "..", "*.py"))
    python_files_to_import = [basename(f)
                              for f in modules if isfile(f) and not f.startswith('__') and f != 'main.py']
    for py_file in python_files_to_import:
        module_name = py_file.removesuffix('.py')
        try:
            importlib.import_module(module_name)
        except Exception as e:
            print(f'ignore {module_name} ({py_file}), error:\n{e}')
            with open(f'{py_file}.error', 'w') as f:
                f.write(str(e))



from collections import defaultdict
import functools
import inspect
import json
import logging
import os
import pickle
from pathlib import Path
from typing import Callable, List, Optional
from unittest.mock import MagicMock

import isort
import ray
from pydantic import BaseModel

from .auto_import import auto_import
from .data_collector import (DataPoint, calculate_candidates_stats_scores,
                             get_candidate_input_output, get_data_collector,
                             get_test_dist, is_data_collection_mode, get_data_dist,
                             reduce_to_most_frequent_answer, datapoint_to_input_output_str)
from .debugger import get_all_test_cases
from .llm import ChatMessage, query_llm
from .python_code import PythonCode
from .round_info import get_round_info
from .test_designer import create_test_file
from .util import (PersistStateToFile, alphanumeric_uuid, entropy_list,
                   load_prompt, make_function_call_statement_str,
                   shorten_answer, shorten_list)
from .data_collector import DATA_DIST



_INPUT = 'input.py'
_OUTPUT = 'output.py'

_REGISTRY = dict()

PROGRAMMER_PROMPT = load_prompt(
    os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "prompts/programmer_prompt_simple.yaml",
    )
)


@ray.remote
def _execute_remote(function_obj, candidate_id, *args, **kwargs):
    print(f'remotely execute {candidate_id}, input={shorten_answer(args)}')
    try:
        ret = function_obj(*args, **kwargs)
    except Exception as e:
        ret = e
    dp = DataPoint(input=args, output=ret, candidate=candidate_id)

    # Check serializable.
    try:
        ray.put(dp)
        return dp
    except:
        pass
    return None


class CodeAgentState(BaseModel, PersistStateToFile):
    function_obj: Callable
    source_code: str
    source_filename: str
    function_name: str

    @property
    def id(self) -> str:
        return f'{self.function_name}({self.source_filename})'

    @property
    def clean_source_code(self) -> str:
        """Without decorator."""
        clean_source_code = self.source_code
        clean_source_code = clean_source_code.replace(
            '@sudocode.CodeAgent()', '')
        clean_source_code = clean_source_code.replace(
            '@sudocode.CodeAgent', '')
        clean_source_code = clean_source_code.replace(
            '@sudocode.DebugAgent()', '')
        clean_source_code = clean_source_code.replace(
            '@sudocode.DebugAgent', '')
        return clean_source_code

    @property
    def module_name(self) -> str:
        return self.source_filename.removesuffix('.py')

    def load_previous_history(self) -> List[ChatMessage]:
        history_filename = Path(f'{self.source_filename}.history.pickle')
        if history_filename.exists():
            return pickle.loads(history_filename.read_bytes())
        return []

    def debug_str(self):
        return self.id

    def debug_oracle_pass(self):
        """Eval if passing against the dataset ground-truth.

        CAUTION: only print in logging to help development. Use this information to code-gen is CHEATING.
        """
        try:
            import eval_output
            eval_output.check(self.function_obj)
        except:
            return False
        return True

    def call_and_record_data_remotely(self, *args, **kwargs):
        return _execute_remote.remote(self.function_obj, self.id, *args, *kwargs)

    def is_genesis(self) -> bool:
        return self.source_filename == _INPUT

    def delete(self):
        logging.info(f'=== Delete {self.id}')
        os.system(f'rm {self.source_filename}*')

    def submit(self):
        final_source_code = f'''# Submitted from {self.id}
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np

from {self.module_name} import *

{self.clean_source_code}

final_answer = {self.function_name}
'''
        with open(_OUTPUT, 'w') as f:
            f.write(final_source_code)

    def query_llm_and_save_candidate(self, message, system_message=PROGRAMMER_PROMPT.format()) -> Optional[List[str]]:
        previous_history = self.load_previous_history()
        new_id = str(alphanumeric_uuid()[-5:])
        output_filename = f'candidate_{new_id}.py'
        replys, histories = query_llm(
            message, system_message=system_message, previous_history=previous_history, filename=output_filename)

        wrote_filenames = []
        for i, reply in enumerate(replys):
            candidate = PythonCode.extract_main_program_code_block_from_full_reply(
                reply, function_name=self.function_name)
            if candidate.valid():
                final_code = f'''# Mutation from '{self.source_filename}'
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np

import sudocode

{candidate.code_text}
'''
                new_candidate_filename = f'candidate_{new_id}_{i}.py'
                with open(new_candidate_filename, 'w') as f:
                    f.write(final_code)
                    logging.info(f'Wrote to {new_candidate_filename}')
                isort.file(new_candidate_filename)
                new_history_filename = f'{new_candidate_filename}.history.pickle'
                with open(new_history_filename, 'wb') as f:
                    pickle.dump(histories[i], f)
                    logging.info(f'Wrote to {new_history_filename}')
                wrote_filenames.append(new_candidate_filename)
        return wrote_filenames

    def generate_test_case(self) -> bool:
        return create_test_file(self.source_filename,
                                self.function_name, self.clean_source_code)

    def generate_seed_candidate(self) -> bool:
        message = f'''
```python
{self.clean_source_code}
```

Generate '{self.function_name}'.
'''
        return bool(self.query_llm_and_save_candidate(message))

    def generate_candidate_by_execution_result(self, execution_results: str) -> bool:
        message = f'''Check the correctness based on execution results.
```
{execution_results}
```

If you found the function {self.function_name} is not correct, re-generate it.
'''
        return bool(self.query_llm_and_save_candidate(message))


def get_all_agents() -> List[CodeAgentState]:
    return [agent for agent in _REGISTRY.values() if not agent.is_genesis()]


def get_genesis() -> CodeAgentState:
    return [agent for agent in _REGISTRY.values() if agent.is_genesis()][0]


def find_agent(candidate_id: str) -> Optional[CodeAgentState]:
    return _REGISTRY.get(candidate_id, None)


def CodeAgent(method=None, **kwargs):
    def _harness(method):
        source_code = inspect.getsource(method)
        source_filename = os.path.basename(method.__code__.co_filename)
        code_agent_state = CodeAgentState(
            function_obj=method, source_code=source_code, source_filename=source_filename, function_name=method.__name__)
        _REGISTRY[code_agent_state.id] = code_agent_state

        @functools.wraps(method)
        def _wrapper(*args, **kwargs):
            # Mock to avoid any test errors (e.g. `assert ==`).
            always_true = MagicMock()
            always_true.__eq__.return_value = True
            always_true.__gt__.return_value = True
            always_true.__lt__.return_value = True

            if is_data_collection_mode():
                call_refs = [agent.call_and_record_data_remotely(
                    *args, **kwargs) for agent in get_all_agents()]
                readys, unreadys = ray.wait(
                    call_refs, num_returns=len(call_refs), timeout=5)
                for task in unreadys:
                    ray.cancel(task, force=True)
                datapoints = ray.get(readys)
                collector = get_data_collector()
                for dp in datapoints:
                    if dp is not None:
                        collector.record(dp)
                return always_true
            else:
                # TODO: non data collection mode
                # return code_agent_state.call_and_maybe_fix_exception(*args, **kwargs)
                return always_true

        setattr(_wrapper, '__code_agent_state__', code_agent_state)
        return _wrapper

    if method is not None:
        return _harness(method)
    return _harness


def print_candidate_rank(ranked_candidate):
    if not ranked_candidate:
        return
    print('\nCandidate rank:')
    for score, candidate_id in ranked_candidate:
        agent = find_agent(candidate_id)
        if agent is not None:
            if agent.debug_oracle_pass():
                oracle_pass = '✔'
            else:
                oracle_pass = '✖'
        else:
            oracle_pass = 'N/A'
        print(f'{score:8}: {candidate_id:50} - {oracle_pass}')


def load_locked_tests():
    if not os.path.exists('locked_tests.json'):
        return {}
    with open('locked_tests.json', 'r') as f:
        return json.load(f)


def save_locked_tests(locked_tests):
    with open('locked_tests.json', 'w') as f:
        json.dump(locked_tests, f, indent=2)


def refresh_all_data():
    DATA_DIST.clear()
    auto_import()
    all_test_cases = get_all_test_cases()
    for test in all_test_cases:
        test()


def generate_code():
    auto_import()
    genesis = get_genesis()
    genesis.generate_seed_candidate()


def generate_test():
    auto_import()
    genesis = get_genesis()
    genesis.generate_test_case()
    refresh_all_data()
    all_test_cases = get_all_test_cases()
    for test in all_test_cases:
        test()


def get_test_and_candidate_info():
    refresh_all_data()
    genesis = get_genesis()
    function_name = genesis.function_name

    locked_tests = load_locked_tests()

    candidate_to_passed_locked_tests = defaultdict(set)
    test_info = []
    candidate_to_input_output = get_candidate_input_output()
    test_dist = get_test_dist()
    for input, dist_with_output in test_dist.items():
        dist = [x[0] for x in dist_with_output]
        outputs = [x[1] for x in dist_with_output]
        most_frequent_output = outputs[0]
        print(
            f'=== {shorten_answer(input)} -> {most_frequent_output}, entropy: {entropy_list(dist):.2f}, dist: {shorten_list(dist)}')
        
        outputs_info = defaultdict(list)
        for candidate_id, input_to_datapoint_dict in candidate_to_input_output.items():
            for _, datapoint in input_to_datapoint_dict.items():
                try:
                    input_str, output_str = datapoint_to_input_output_str(datapoint)
                except:
                    continue
                if input_str != input:
                    continue
                
                call_str = make_function_call_statement_str(datapoint.input, datapoint.output, function_name, limit=1000)
                outputs_info[output_str].append({
                    'datapoint': datapoint,
                    'call_str': call_str,
                    'candidate_id': candidate_id,
                })

        default_output_str = most_frequent_output
        locked_output = locked_tests.get(input, None)
        locked = False
        if (locked_output is not None) and (locked_output in outputs):
            default_output_str = locked_output
            locked = True

        this_test_id = alphanumeric_uuid()
        if locked:
            correct_info_list = outputs_info[default_output_str]
            for info in correct_info_list:
                candidate_to_passed_locked_tests[info['candidate_id']].add(this_test_id)

        test_info.append({
            'id': this_test_id,
            'input': input,
            'default_output_str': default_output_str,
            'default_call_str': outputs_info[default_output_str][0]['call_str'],
            'outputs': outputs,  # in rank order.
            'outputs_info': outputs_info,
            'locked': locked,
        })

    all_candidates = get_all_agents()
    all_candidate_ids = [c.id for c in all_candidates]

    candidate_to_stats_scores = calculate_candidates_stats_scores(all_candidate_ids)
    candidate_info = [{
            'candidate_id': candidate_id,
            'candidate': find_agent(candidate_id),
            'stats_score': stats_score,
            'tests_score': len(candidate_to_passed_locked_tests[candidate_id]),
            'passed_tests': candidate_to_passed_locked_tests[candidate_id],
        } for candidate_id, stats_score in candidate_to_stats_scores.items()]
    
    test_info = sorted(test_info, key=lambda x:x['locked'], reverse=True)
    candidate_info = sorted(candidate_info, key=lambda x:(x['tests_score'], x['stats_score']), reverse=True)
    return test_info, candidate_info, locked_tests



import geniz
from input import make_palindrome


@geniz.DebugAgent()
def test_make_palindrome():
    assert make_palindrome('') == ''
    assert make_palindrome('cat') == 'catac'
    assert make_palindrome('tac') == 'tac'
    assert make_palindrome('abba') == 'abbaab'
    assert make_palindrome('abcd') == 'dccba'
    assert make_palindrome('abccba') == 'abccbac'
    assert make_palindrome('race') == 'ecarace'
    assert make_palindrome('programming') == 'mnangrommargopgnir'
    assert make_palindrome('noon') == 'noonno'
    assert make_palindrome('cyclist') == 'ticsyrclitcy'
    assert make_palindrome('123454321') == '1234532431'
    assert make_palindrome('education') == 'noitoracedu'
    assert make_palindrome('a') == 'a'
    assert make_palindrome('ab') == 'baab'
    assert make_palindrome('break') == 'ekiribarb'
    assert make_palindrome('ballet') == 'elttalabeltabel'
    assert make_palindrome('abcdef') == 'fedcbaabcdef'
    assert make_palindrome('aabsnrvsd') == 'vsdrnsabarsn'
    print('All test cases passed!')
from input import make_palindrome

test_make_palindrome()



import os
import litellm
from typing import Dict, List, Optional

from pydantic import BaseModel

from .round_info import get_round_info
from .keys import MODEL, API_BASE, API_KEY

N = 1


class ChatMessage(BaseModel):
    role: str
    content: str

    def to_dict(self) -> Dict[str, str]:
        return {"role": self.role, "content": self.content}

    def to_string(self) -> str:
        return f"{self.role}: {self.content}"


def query_llm(prompt: str, *, system_message: str = '', previous_history: List[ChatMessage] = [], filename: Optional[str] = None, model=MODEL, n=N):
    messages: List[ChatMessage] = (
        previous_history + [ChatMessage(role='user', content=prompt)])
    if filename is not None:
        with open(f'{filename}.prompt.txt', 'w') as f:
            f.write(f'{model}\n\n{system_message}\n\n')
            for i, message in enumerate(messages):
                f.write(f'=== {i}: {message.role} ===\n')
                f.write(message.content)
                f.write('\n')

    final_messages = [msg.to_dict() for msg in messages]
    if system_message:
        final_messages = [
            {"role": "system", "content": system_message}] + final_messages
        
    round_info = get_round_info()
    session_id = round_info.session_id

    if n == 1:
        n = None
    response = litellm.completion(
        model=f"openai/{model}",
        api_key=API_KEY,
        api_base=API_BASE,
        messages=final_messages,
        temperature=1.2,
        num_retries=1,
        n=n,
        metadata={
            "generation_name": "autogenesis-0510",
            "generation_id": "autogenesis-0510",
            "version":  "0510",
            "trace_user_id": "ning",
            "session_id": session_id,
            "tags": ["autogenesis", MODEL],
        },
    )
    replys = [choice.message.content for choice in response.choices]
    histories = []
    for reply in replys:
        new_messages = messages + \
            [ChatMessage(role='assistant', content=reply)]
        histories.append(new_messages)
    if filename is not None:
        with open(f'{filename}.prompt.txt', 'a') as f:
            for i, reply in enumerate(replys):
                f.write(f'\n=== Reply {i} ===\n')
                f.write(reply)
    return replys, histories



import argparse
import json
import os
import shutil
import traceback
import uuid
from pathlib import Path
from time import sleep
from typing import List, Optional

import docker
from docker.errors import NotFound
from docker.models.containers import Container

_IMAGE = 'autogenesis'


def _wait_for_ready(container: Container, timeout: int = 60, stop_time: float = 0.1) -> None:
    elapsed_time = 0
    while container.status != "running" and elapsed_time < timeout:
        sleep(stop_time)
        elapsed_time += stop_time
        container.reload()
        continue
    if container.status != "running":
        raise ValueError("Container failed to start")


def _attach_or_create_container(container_name: str, world_folder: Path, docker_image: str = _IMAGE, model: str = 'Phi-3-mini-128k-instruct') -> Container:
    client = docker.from_env()
    try:
        container = client.containers.get(container_name)
        return container
    except NotFound:
        container = client.containers.create(
            docker_image,
            name=container_name,
            entrypoint="/bin/sh",
            tty=True,
            auto_remove=True,
            volumes={
                '/var/run/docker.sock': {'bind': '/var/run/docker.sock', 'mode': 'rw'},
                str(world_folder.resolve()): {"bind": "/workspace", "mode": "rw"}
            },
            network_mode='host',
            working_dir="/workspace",
        )
        container.start()
        return container


def _stop_container(container_name: str):
    client = docker.from_env()
    try:
        container = client.containers.get(container_name)
        container.stop()
    except:
        return


def stop(world_folder: Path):
    docker_name_file = Path(os.path.join(world_folder, 'container_name.txt'))
    if docker_name_file.exists():
        container_name = docker_name_file.read_text()
        _stop_container(container_name)


def init_world_folder(world_folder: Path, source_agent_folder: Path):
    """init world folder with a pre-configuared example."""
    shutil.copytree(source_agent_folder, world_folder)


def run(world_folder: Path, cmd: List[str] = ['python', 'main.py'], timeout: float = 1200, save_log_to_file: bool = False, image: str = _IMAGE, model: str = 'Phi-3-mini-128k-instruct'):
    docker_name_file = Path(os.path.join(world_folder, 'container_name.txt'))
    task_id = Path(os.path.join(world_folder, 'task_id.txt'))
    log_file = Path(os.path.join(world_folder, 'stdout.log'))

    if not docker_name_file.exists():
        if task_id.exists():
            normalized_task_id = task_id.read_text().replace('/', '-')
            container_name = f'autogenesis-{normalized_task_id}-{str(uuid.uuid4())[-5:]}'
        else:
            container_name = f'autogenesis-{uuid.uuid4()}'
        docker_name_file.write_text(container_name)
    container_name = docker_name_file.read_text()

    container = _attach_or_create_container(
        container_name, world_folder, image, model)
    _wait_for_ready(container)

    command = ["timeout", str(timeout)] + cmd
    result = container.exec_run(command, stream=True, environment={'AUTOGENESIS_MODEL': model})
    output_stream = result.output
    for chunk in output_stream:
        if save_log_to_file:
            with open(log_file, 'ab') as f:
                f.write(chunk)
        print(chunk.decode('utf-8'), flush=True)
    return result


def check_stop_criteria(world_folder: Path, rounds: Optional[int] = None, save_log_to_file: bool = False, image: str = _IMAGE):
    submitted_file = Path(os.path.join(world_folder, 'output.py'))
    pass_file = Path(os.path.join(world_folder, 'output.pass'))
    if submitted_file.exists():
        run(world_folder, cmd=['python', 'eval_output.py'],
            save_log_to_file=save_log_to_file, image=image)
        if pass_file.exists():
            passed = pass_file.read_text()
            print(f'{world_folder} passed: {passed}')
        else:
            stop(world_folder)
            dirname, basename = os.path.split(world_folder)
            dest_path = os.path.join(dirname, f'{basename}.abnormal')
            shutil.rmtree(dest_path, ignore_errors=True)
            world_folder.replace(dest_path)
            print(f'{world_folder} abnormal')
        return True

    if rounds is not None:
        roundinfo_file = Path(os.path.join(world_folder, 'RoundInfo.json'))
        if roundinfo_file.exists():
            roundinfo = json.loads(roundinfo_file.read_text())
            current_round = int(roundinfo['round'])
            if current_round >= rounds:
                print(f'Reach max rounds {rounds}, stop')
                return True
    return False


def main():
    parser = argparse.ArgumentParser(description="Autogenesis Runner")
    parser.add_argument("--world", type=str, default="./world_folder",
                        help="The world folder.")
    parser.add_argument("--agent", type=str, default="./agentcoder_v0",
                        help="Source agent.")
    parser.add_argument("--image", type=str, default="autogenesis",
                        help="Docker image")
    parser.add_argument("--model", type=str, default="Phi-3-mini-128k-instruct",
                        help="LLM model")
    parser.add_argument("--rounds", type=int, default=1, help="Max rounds")
    parser.add_argument("--save_log_to_file", action='store_true',
                        help="Also save terminal log to 'stdout.log' in the world folder.")
    parser.add_argument("--init_only", action='store_true',
                        help="Only init the world folder without run.")
    args = parser.parse_args()

    world_folder_path = Path(args.world)
    if not world_folder_path.exists():
        try:
            init_world_folder(world_folder_path, Path(args.agent))
        except:
            print('Init failed.')
            traceback.print_exc()
            shutil.rmtree(world_folder_path, ignore_errors=True)
            exit(1)

    if args.init_only:
        exit(0)

    for _ in range(args.rounds):
        try:
            run(world_folder_path,
                save_log_to_file=args.save_log_to_file,
                image=args.image,
                model=args.model)
            if check_stop_criteria(world_folder_path, args.rounds, save_log_to_file=args.save_log_to_file, image=args.image):
                break
        except:
            traceback.print_exc()
            break
    stop(world_folder_path)


if __name__ == '__main__':
    main()




import json
import string
import uuid
from math import log2
from pathlib import Path
from typing import Any, List

import yaml
from jinja2 import Template
from pydantic import BaseModel


class PersistStateToFile:

    @classmethod
    def filename(cls):
        return f'{cls.__name__}.json'

    @classmethod
    def load(cls):
        try:
            with open(cls.filename(), 'r') as f:
                json_dict = json.load(f)
                return cls(**json_dict)
        except:
            return cls()

    def save(self):
        with open(self.filename(), 'w') as f:
            f.write(self.model_dump_json(indent=2))


BASE62 = string.ascii_uppercase + string.ascii_lowercase + string.digits


def base62_encode(num) -> str:
    if num == 0:
        return BASE62[0]
    arr = []
    while num:
        num, rem = divmod(num, 62)
        arr.append(BASE62[rem])
    arr.reverse()
    return "".join(arr)


def alphanumeric_uuid() -> str:
    """Generate a base62-encoded UUID for IDs that need to be URL-safe."""
    uuid_hex = uuid.uuid4().hex
    uuid_int = int(uuid_hex, 16)
    return base62_encode(uuid_int)


def shorten_answer(var: Any, limit: int = 50) -> str:
    var_s = str(var)
    if len(var_s) > limit:
        var_s = var_s[:limit] + '...'
    return var_s


def shorten_list(var: List, limit: int = 5) -> str:
    if len(var) > limit:
        return str(var[:limit] + ['...'])
    return str(var)


def entropy_list(var: List) -> float:
    def normalize(data):
        """
        Normalize an array of integers to get probabilities.
        """
        data_sum = sum(data)
        probabilities = [val / data_sum for val in data]
        return probabilities

    def entropy(probabilities):
        """
        Calculate the entropy of a list of probabilities.
        """
        ent = 0.0
        for prob in probabilities:
            if prob > 0:
                ent += -prob * log2(prob)
        return ent

    return entropy(normalize(var))


def make_function_call_statement_str(input, output, function_name, limit=200) -> str:
    inputs = []
    for i in input:
        if isinstance(i, str):
            inputs.append(f'"{i}"')
        else:
            inputs.append(str(i))
    if not inputs:
        return ''
    args_str = ', '.join(inputs)

    if isinstance(output, str):
        output_str = f'"{output}"'
    elif isinstance(output, Exception):
        output_str = f'Exception("{str(output)}")'
    else:
        output_str = str(output)

    formatted_str = f'{function_name}({args_str}) -> {output_str}'
    if limit:
        return shorten_answer(formatted_str, limit=limit)
    return formatted_str


class PromptTemplate(BaseModel):
    """Lightweight prompt template implementation."""

    template: str
    input_variables: List[str]
    template_format: str  # ["f-string", "jinja2"]

    def format(self, **kwargs):
        """Format the prompt template."""
        if self.template_format == "jinja2":
            return self._format_jinja2(**kwargs)
        return self.template.format(**kwargs)

    def _format_jinja2(self, **kwargs):
        return Template(self.template).render(**kwargs)


def load_prompt(path: str) -> PromptTemplate:
    """Load prompt from file."""
    file_path = Path(path)
    if file_path.suffix == ".json":
        with open(file_path) as f:
            config = json.load(f)
    elif file_path.suffix == ".yaml":
        with open(file_path, "r") as f:
            config = yaml.safe_load(f)
    else:
        raise ValueError(f"Got unsupported file type {file_path.suffix}")

    return PromptTemplate(
        template=config["template"],
        input_variables=config["input_variables"],
        template_format=(
            config["template_format"] if "template_format" in config else "f-string"
        ),
    )



# Mutation from 'input.py'
import os
import sys
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np

import geniz


@geniz.CodeAgent()
def make_palindrome(s: str) -> str:
    """
    Find the shortest palindrome that begins with a supplied string.
    The goal is to append the reverse of the longest possible prefix
    (before the identified palindromic suffix) at the end to create a palindrome.
    
    Args:
    s (str): The input string.

    Returns:
    str: The shortest palindrome that begins with the provided string.
    
    Examples:
    >>> make_palindrome('')
    ''
    >>> make_palindrome('cat')
    'catac'
    >>> make_palindrome('cata')
    'catac'
    """
    if not s:
        return s
    for i in range(len(s), -1, -1):
        if s[:i] * (len(s) // i + 1) == s[:i] * -(-len(s) // i):
            suffix = s[:i]
            return s + suffix[::-1]
    return s + s[::-1]
if __name__ == '__main__':
    print(make_palindrome(''))
    print(make_palindrome('cat'))
    print(make_palindrome('cata'))
    print(make_palindrome('hello'))



import functools
import os

from .data_collector import data_collection_mode, is_data_collection_mode

_REGISTRY = []


def DebugAgent(**kwargs):
    def _harness(method):
        @functools.wraps(method)
        def _wrapper(*args, **kwargs):
            name = method.__name__
            source_filename = os.path.basename(method.__code__.co_filename)
            test_case = f'{name}({source_filename})'
            if is_data_collection_mode():
                print(
                    f'>>> Skip test case {test_case}, nested data collection mode')
                return
            print(
                f'>>> Run test case {test_case} in data collection mode')
            with data_collection_mode(test_case=test_case):
                method(*args, **kwargs)

        _REGISTRY.append(_wrapper)
        return _wrapper

    return _harness


def get_all_test_cases():
    return _REGISTRY
